      subroutine htrns(tmp,u,v,sch,dx,dz,dxs,dzs,asp,
     $     tsurf,tbot,qbot,vboti,dt,jstas,jends)
      use para4common
***********************************************************************
*     solve energy equation by implicit, power-law scheme
*     adopted boundary conditions are;
*        fixed temperature on top/bottom boundaries
*        zero heat flux across side walls
*
*     discretized linear equation for temperature is solved by
*     red-black SOR (Successive Over-Relaxation) method
*
*     ver.0 on 20021119; red-black SOR is implemented (for 1cpu only)
*     ver.1 in 20021202; parallelized
*
*     input
*        tmp: temperature at time step n
*        u,v: velocity
*        sch: heat generation
*        tsurf: surface temperature
*        tbot: bottom temperature at time step n
*        vboti: inverse of volume of heat bath at bottom
*        dx,dz: mesh size
*        dt: time step
*     output
*        tmp: temperature at time step n+1
*        tbot: bottom temperature at time step n+1
*        qbot: bottom heat flux
*     local
*        pe: Peclet number
*        twk: work array for temperature
*        fx,fz: heat flux in x/z direction
**********************************************************************
      implicit none
      include "meshnum"
c
      integer i,j
      real*8 tmp(iax,jax),u(iax,jax),v(iax,jax),sch(iax,jax)
      real*8 dx(iax),dz(jax),dxs(iax),dzs(jax),asp,dt
      real*8 twk(iax,jax)
      real*8 tsurf,tbot,qbot,vboti,tb
      real*8 difx(iax,jax),difz(iax,jax)
      real*8 pe,faca,facb
      real*8 cax(iax,jax),caz(iax,jax),cbx(iax,jax),cbz(iax,jax)
      real*8 dum,at1(5,iax,jax),bt1(iax,jax)
c
      integer jstas(0:maxprocs-1),jends(0:maxprocs-1)
      integer jsta,jend,jsta2,jend1
c
c     variables for linear solver
c
      integer nitr,maxsor,istat
      parameter(maxsor=9999)
      real*8 relt,eitr,eps
      parameter(eps=1.d-10,relt=1.d0)
c
      jsta = jstas(myrank)
      jend = jends(myrank)
      jsta2 = max(2,jsta)
      jend1 = min(jax-1,jend)
c
c ---------------------------------------
c     temperature change in heat bath
c ---------------------------------------
c
      if (myrank.eq.0) then
         tb = 0.d0
         do i = 2,iax-1
            tb = tb + tmp(i,2)*dx(i)
         enddo
         tb = tb/asp
      endif
      call mpi_bcast(tb,1,mpi_double_precision,0,mpi_comm_world,ierr)
      qbot = 2.d0*(tbot-tb)/dz(2)
      tbot = tbot - vboti*qbot*dt
c
c -----------------------
c     set work arrays
c -----------------------
c
c      do j = 2,jax-1
      do j = jsta2,jend1
         do i = 2,iax-1
            twk(i,j) = tmp(i,j)
         enddo
      enddo
c
c     top/bottom boundary
c
      if (myrank.eq.0) then
         do i = 2,iax-1
            twk(i,1) = (2.d0*dzs(2)*tbot-dz(1)*tmp(i,2))/dz(2)
         enddo
      endif
      if (myrank.eq.nprocs-1) then
         do i = 2,iax-1
            twk(i,jax) = (2.d0*dzs(jax)*tsurf-dz(jax)*tmp(i,jax-1))
     $           /dz(jax-1)
         enddo
      endif
c
c     side walls
c
      do j = jsta,jend
         twk(1,j) = tmp(2,j)
         twk(iax,j) = tmp(iax-1,j)
      enddo
c
c -------------------
c     conductance
c -------------------
c
c      do j = 2,jax-1
      do j = jsta2,jend1
         do i = 2,iax
            difx(i,j) = 1.d0/dxs(i)
         enddo
      enddo
c
c      do j = 2,jax
      do j = jsta2,jend
         do i = 2,iax-1
            difz(i,j) = 1.d0/dzs(j)
         enddo
      enddo
c
c ----------------------------------------
c     calculate heat flux coefficients
c ----------------------------------------
c
c      do j = 2,jax-1
      do j = jsta2,jend1
         do i = 2,iax
            pe = u(i,j)/difx(i,j)
            dum = 1.d0-0.1d0*dabs(pe)
            faca = dmax1(0.d0,dum**5)+dmax1(0.d0,-pe)
            facb = faca + pe
            cax(i,j) = difx(i,j)*faca
            cbx(i,j) = difx(i,j)*facb
         enddo
      enddo
c
c      do j = 2,jax
      do j = jsta2,jend
         do i = 2,iax-1
            pe = v(i,j)/difz(i,j)
            dum = 1.d0-0.1d0*dabs(pe)
            faca = dmax1(0.d0,dum**5)+dmax1(0.d0,-pe)
            facb = faca + pe
            caz(i,j) = difz(i,j)*faca
            cbz(i,j) = difz(i,j)*facb
         enddo
      enddo
      call mypartialshift(caz,jsta,jend,2)
      call mypartialshift(cbz,jsta,jend,2)
c
c ----------------------------------------------
c     calculate coefficients for temperature
c ----------------------------------------------
c
c      do j = 2,jax-1
      do j = jsta2,jend1
         do i = 2,iax-1
            at1(1,i,j) = -cbx(i,j)*dt/dx(i)
            at1(2,i,j) = -cbz(i,j)*dt/dz(j)
            at1(4,i,j) = -caz(i,j+1)*dt/dz(j)
            at1(5,i,j) = -cax(i+1,j)*dt/dx(i)
            at1(3,i,j) = 1.d0 + dt*(
     $           (cax(i,j)+cbx(i+1,j))/dx(i)
     $           +(caz(i,j)+cbz(i,j+1))/dz(j))
            bt1(i,j) = twk(i,j) + sch(i,j)*dt
         enddo
      enddo
c
c ---------------------------
c     boundary conditions
c ---------------------------
c
      if (myrank.eq.0) then
         do i = 2,iax-1
            at1(3,i,2) = at1(3,i,2)
     $           - dz(1)/dz(2)*at1(2,i,2)
            bt1(i,2) = bt1(i,2)
     $           - 2.d0*dzs(2)/dz(2)*at1(2,i,2)*tbot
            at1(2,i,2) = 0.d0
         enddo
      endif
      if (myrank.eq.nprocs-1) then
         do i = 2,iax-1
            at1(3,i,jax-1) = at1(3,i,jax-1)
     $           - dz(jax)/dz(jax-1)*at1(4,i,jax-1)
            bt1(i,jax-1) = bt1(i,jax-1)
     $           - 2.d0*dzs(jax)/dz(jax-1)*at1(4,i,jax-1)*tsurf
            at1(4,i,jax-1) = 0.d0
         enddo
      endif
      do j = jsta2,jend1
         at1(3,2,j) = at1(3,2,j) + at1(1,2,j)
         at1(1,2,j) = 0.d0
      enddo
      do j = jsta2,jend1
         at1(3,iax-1,j) = at1(3,iax-1,j) + at1(5,iax-1,j)
         at1(5,iax-1,j) = 0.d0
      enddo
c
c /////////////////////////////////////
c     begin linear solver
c     by Successive Over-Relaxation
c /////////////////////////////////////
c
      nitr = maxsor
      eitr = eps
      call sor2c2p(at1,bt1,twk,relt,istat,nitr,eitr,jstas,jends)
c
      if (myrank.eq.0) then
         if (istat.eq.0) then
            write(*,'("OK/SOR ",i4,e11.4)')nitr,eitr
         else
            write(*,'("NG/SOR ",i4,e11.4)')nitr,eitr
         endif
      endif
c      call myooban(twk,jstas,jends)
c
c ====================
c     SOR loop end
c ====================
c
      do j = jsta2,jend1
         do i = 2,iax-1
            tmp(i,j) = twk(i,j)
         enddo
      enddo
c
c     top/bottom boundary
c
      if (myrank.eq.0) then
         do i = 2,iax-1
            tmp(i,1) = (2.d0*dzs(2)*tbot-dz(1)*tmp(i,2))/dz(2)
         enddo
      endif
      if (myrank.eq.nprocs-1) then
         do i = 2,iax-1
            tmp(i,jax) = (2.d0*dzs(jax)*tsurf-dz(jax)*tmp(i,jax-1))
     $           /dz(jax-1)
         enddo
      endif
c
c     side walls
c
      do j = jsta,jend
         tmp(1,j) = tmp(2,j)
         tmp(iax,j) = tmp(iax-1,j)
      enddo
c
      return
      end
