      subroutine ssorcg(a,b,x,r,p,ru,ap,mx,mz,eps,nitr)
**********************************************************************
*     conjugate gradient method for SSOR-CG
*     solve 2-D Poisson equation by FDM
*
*     ver.0 on 20021119
*
*     global variables:
*       mx,mz: number of solution points in each direction
*       a(5,mx+2,mz+2): nonzero elements of coefficient matrix
*       b(mx+2,mz+2): RHS vector
*       x(mx+2,mz+2): initial guess on input, solution on output
*       eps: tolerance on input, truncation error on output
*       nitr: maximum iteration on input, iteration number on output
*       r(mx+2,mz+2):
*       p(mx+2,mz+2):
*       ru(mx+2,mz+2):
*       ap(mx+2,mz+2):
*     arrays for message passing
*       is1,is2,ir1,ir2,bufs1,bufs2,bufr1,bufr2
**********************************************************************
      implicit none
c
      integer i,j,k,ncg,maxitr
      integer mx,mz,nitr
      real*8 a(5,mx+2,mz+2),b(mx+2,mz+2),x(mx+2,mz+2)
      real*8 r(mx+2,mz+2),p(mx+2,mz+2),ru(mx+2,mz+2)
     $     ,ap(mx+2,mz+2)
      real*8 alpha,b2,beta,eps,eps2
      real*8 rur0,rur1,pap,res2
c
      maxitr = nitr
      eps2 = eps**2
c
c ----------------------
c     initialization
c ----------------------
c
      call prodmv(a,x,ap,mx,mz)
c
      do j = 2,mz+1
         do i = 2,mx+1
            r(i,j) = b(i,j)-ap(i,j)
         enddo
      enddo
c
      call ssorsl(a,r,ru,mx,mz)
c
      do j = 2,mz+1
         do i = 2,mx+1
            p(i,j) = ru(i,j)
         enddo
      enddo
      call prodvv(r,ru,rur0,mx,mz)
c
      call prodvv(b,b,b2,mx,mz)
c
c ============================
c     iterative loop start
c ============================
c
      do ncg = 1,maxitr
c
         call prodmv(a,p,ap,mx,mz)
c
         call prodvv(p,ap,pap,mx,mz)
c
         alpha = rur0/pap
         do j = 2,mz+1
            do i = 2,mx+1
               x(i,j) = x(i,j)+alpha*p(i,j)
               r(i,j) = r(i,j)-alpha*ap(i,j)
            enddo
         enddo
         call prodvv(r,r,res2,mx,mz)
c
         eps = dsqrt(res2/b2)
         if (res2/b2.le.eps2) then
c
c -----------------
c     converged
c -----------------
c
            nitr = ncg
            goto 1000
         endif
c
c ---------------------
c     not converged
c ---------------------
c
         call ssorsl(a,r,ru,mx,mz)
c
         call prodvv(r,ru,rur1,mx,mz)
         beta = rur1/rur0
         rur0 = rur1
c
         do j = 2,mz+1
            do i = 2,mx+1
               p(i,j) = ru(i,j)+beta*p(i,j)
            enddo
         enddo
c
c ==========================
c     iterative loop end
c ==========================
c
      enddo
c
 1000 continue
c
      return
      end
c
c --------------------------------------------------------------------
c
      subroutine prodmv(a,x,ap,mx,mz)
**********************************************************************
*     Compute product
*     AP = (matrix A) * (vector X)
**********************************************************************
      implicit none
c
      integer i,j,k,ij
      integer mx,mz
      real*8 a(5,mx+2,mz+2),x(mx+2,mz+2),ap(mx+2,mz+2)
c
      do j = 2,mz+1
         do i = 2,mx+1
            ap(i,j) = a(1,i,j)*x(i,j-1)+a(2,i,j)*x(i-1,j)
     $           +a(3,i,j)*x(i,j)+a(4,i,j)*x(i+1,j)+a(5,i,j)*x(i,j+1)
         enddo
      enddo
c
      return
      end
c
c --------------------------------------------------------------------
c
      subroutine prodvv(vec1,vec2,vdotv,mx,mz)
**********************************************************************
*     Compute dot product of two vectors
*     vdotv = (vec1,vec2)
**********************************************************************
      implicit none
c
      integer i,j,k
      integer mx,mz
      real*8 vec1(mx+2,mz+2),vec2(mx+2,mz+2),vdotv
c
      real*8 vdotvtmp
c
      vdotvtmp = 0.d0
      do j = 2,mz+1
         do i = 2,mx+1
            vdotvtmp = vdotvtmp+vec1(i,j)*vec2(i,j)
         enddo
      enddo
c
      vdotv = vdotvtmp
c
      return
      end
c
c --------------------------------------------------------------------
c
      subroutine ssorsl(a,b,x,mx,mz)
**********************************************************************
*     preconditioner for CG solving 3-D Poisson equation by FDM
*     Symmetric Successive Over Relaxation method by 2 coloring
*
*     ver.0 on 20020823; as stand-alone SSOR solver for Ax=b
*
*     global variables:
*       a: non-zero elements of coefficients
*       b: elements of right-hand side vector
*       x: initial guess on input, solution on output
*       mx,my,mz: number of solution points in each direction
*     local variables
*       maxsor: maximun iteration number
*       eps: maximum tolerance
*       relc: relaxation coefficient
*       eitr,eitrtmp: residual
**********************************************************************
      implicit none
c
      integer i,j,k,ij
      integer mx,my,mz
      real*8 a(5,mx+2,mz+2),b(mx+2,mz+2),x(mx+2,mz+2)
c
      integer maxsor,nsor
      parameter(maxsor=1)
      real*8 eitr,eps,relc,eitrtmp,zansa
      parameter(eps=1.d-10,relc=1.d0)
c
c     initial guess of solution
c
      do j = 2,mz+1
         do i = 2,mx+1
            x(i,j) = 0.d0
         enddo
      enddo
c
c ==================
c     loop start
c ==================
c
      do nsor = 1,maxsor
         eitrtmp = 0.d0
c
c +++++++++++++++++++
c     forward SOR
c +++++++++++++++++++
c
c     1st color (for even i+j)
c
         do j = 2,mz+1
            do i = mod(j,2)+2,mx+1,2
               zansa = (b(i,j)
     $              -a(1,i,j)*x(i,j-1)-a(2,i,j)*x(i-1,j)
     $              -a(4,i,j)*x(i+1,j)-a(5,i,j)*x(i,j+1)
     $              )/a(3,i,j)-x(i,j)
               x(i,j) = x(i,j)+relc*zansa
               eitrtmp = dmax1(eitrtmp,dabs(zansa))
            enddo
         enddo
c
c     2nd color (for odd i+j)
c
         do j = 2,mz+1
            do i = mod(j+1,2)+2,mx+1,2
               zansa = (b(i,j)
     $              -a(1,i,j)*x(i,j-1)-a(2,i,j)*x(i-1,j)
     $              -a(4,i,j)*x(i+1,j)-a(5,i,j)*x(i,j+1)
     $              )/a(3,i,j)-x(i,j)
               x(i,j) = x(i,j)+relc*zansa
               eitrtmp = dmax1(eitrtmp,dabs(zansa))
            enddo
         enddo
c
c ++++++++++++++++++++
c     backward SOR
c ++++++++++++++++++++
c
c     2nd color (for odd i+j)
c
         do j = mz+1,2,-1
            do i = mx+mod(mx+1+j,2),2,-2
               zansa = (b(i,j)
     $              -a(1,i,j)*x(i,j-1)-a(2,i,j)*x(i-1,j)
     $              -a(4,i,j)*x(i+1,j)-a(5,i,j)*x(i,j+1)
     $              )/a(3,i,j)-x(i,j)
               x(i,j) = x(i,j)+relc*zansa
               eitrtmp = dmax1(eitrtmp,dabs(zansa))
            enddo
         enddo
c
c     1st color (for even i+j)
c
         do j = mz+1,2,-1
            do i = mx+mod(mx+j,2),2,-2
               zansa = (b(i,j)
     $              -a(1,i,j)*x(i,j-1)-a(2,i,j)*x(i-1,j)
     $              -a(4,i,j)*x(i+1,j)-a(5,i,j)*x(i,j+1)
     $              )/a(3,i,j)-x(i,j)
               x(i,j) = x(i,j)+relc*zansa
               eitrtmp = dmax1(eitrtmp,dabs(zansa))
            enddo
         enddo
c
         eitr = eitrtmp
c
c         if (eitr.lt.eps) then
c            write(*,'("OK/SSOR ",i4,e11.4)')nsor,eitr
c            goto 999
c         endif
      enddo
c      write(*,'("NG/SSOR ",i4,e11.4)')maxsor,eitr
 999  continue
c
c ================
c     loop end
c ================
c
      return
      end
